import json
from typing import Any

from communica.utils import ByteSeq
from communica.exceptions import FeatureNotAvailable

from .base import BaseSerializer


class JsonSerializer(BaseSerializer):
    """
    This serializer just passes data to json dump function.
    """
    __slots__ = ()

    def json_dumpb(self, obj: Any) -> bytes:
        return json.dumps(obj, ensure_ascii=False).encode('utf-8')

    def json_loadb(self, __obj: ByteSeq) -> Any:
        if not isinstance(__obj, (bytes, str)):
            __obj = bytes(__obj)
        return json.loads(__obj)

    def dump(self, data: Any) -> bytes:
        return self.json_dumpb(data)

    def load(self, raw_data: ByteSeq) -> Any:
        return self.json_loadb(raw_data)


try:
    import orjson

    class OrjsonSerializer(JsonSerializer):  # pyright: ignore[reportRedeclaration]
        __slots__ = ('_option',)

        def __init__(
                self,
                orjson_dump_option: int = orjson.OPT_NON_STR_KEYS
        ) -> None:
            self._option = orjson_dump_option

        def orjson_dumpb(self, obj: Any) -> bytes:
            return orjson.dumps(obj, option=self._option)

        orjson_loadb = orjson.loads

    json_dumpb = OrjsonSerializer().json_dumpb
    json_loadb = OrjsonSerializer().json_loadb
except ModuleNotFoundError:
    class OrjsonSerializer(JsonSerializer):
        def __init__(
                self,
                orjson_dump_option: int = orjson.OPT_NON_STR_KEYS
        ) -> None:
            raise FeatureNotAvailable(
                'OrjsonSerializer requires orjson library. '
                'Install communica with [orjson] extra.'
            )

    json_dumpb = JsonSerializer().json_dumpb
    json_loadb = JsonSerializer().json_loadb



import json
import asyncio
import logging
from abc import ABC, abstractmethod
from ssl import SSLContext
from struct import Struct
from typing import Any, Dict, Type, ClassVar, Optional
from contextlib import suppress
from dataclasses import dataclass

from typing_extensions import Self

from communica.utils import (
    ETX_CHAR,
    NULL_CHAR,
    UINT32MAX,
    MessageQueue,
    fmt_task_name,
    read_accessor,
)
from communica.exceptions import FeatureNotAvailable
from communica.connectors.base import (
    Handshaker,
    HandshakeOk,
    BaseConnector,
    HandshakeFail,
    BaseConnection,
    ClientConnectedCB,
    RequestReceivedCB,
    BaseConnectorServer,
)
from communica.serializers.json import json_dumpb, json_loadb
from communica.connectors.stream import local_connections


__all__ = (
    'TcpConnector',
    'LocalConnector'
)

logger = logging.getLogger('communica.connectors.stream')
DEFAULT_STREAM_HWM = 32


class Frame(ABC):
    CODE: ClassVar[int]
    code_byte: ClassVar[bytes]
    LENGTH_PACKER = Struct(
        '!I'  # big-endian uint32
    )

    length_size = LENGTH_PACKER.size
    length_pack = LENGTH_PACKER.pack
    length_unpack_from = LENGTH_PACKER.unpack_from

    _frames: ClassVar[Dict[int, Type[Self]]] = {}

    def __init_subclass__(cls) -> None:
        cls._frames[cls.CODE] = cls
        cls.code_byte = bytes([cls.CODE])

    @abstractmethod
    def to_bytes(self) -> bytes:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _load(cls, data: memoryview):
        raise NotImplementedError

    @classmethod
    def from_bytes(cls, data: memoryview) -> Self:
        try:
            frame = cls._frames[data[0]]
        except ValueError:
            raise ValueError('Unknown protocol') from None

        return frame._load(data[1:])


@dataclass
class MessageFrame(Frame):
    CODE = 10

    metadata: Any
    raw_data: 'bytes | memoryview'

    @classmethod
    def _load(cls, data: memoryview):
        metadata_end = cls.length_unpack_from(data)[0]
        return cls(
            json_loadb(data[cls.length_size:metadata_end]),
            data[metadata_end:]
        )

    def to_bytes(self) -> bytes:
        return (
            self.code_byte +
            self.length_pack(len(self.metadata) + self.length_size) +
            self.metadata +
            self.raw_data
        )


@dataclass
class CloseNotifyFrame(Frame):
    CODE = 20

    @classmethod
    def to_bytes_with_header(cls) -> bytes:
        return cls.length_pack(len(cls.code_byte)) + cls.code_byte

    @classmethod
    def _load(cls, data: memoryview):
        return cls()

    def to_bytes(self) -> bytes:
        return self.code_byte


class StreamConnection(BaseConnection):
    __slots__ = ('reader', 'writer', 'connector', '_send_queue')

    reader: asyncio.StreamReader
    writer: asyncio.StreamWriter
    connector: 'BaseStreamConnector'

    _send_queue: 'MessageQueue[Frame]'
    _handshake_result: 'HandshakeOk | None'

    _MAX_CHUNK_SIZE = UINT32MAX + 1 - Frame.length_size

    max_chunk_size: read_accessor[int] = read_accessor('_MAX_CHUNK_SIZE')
    """Max size of one message in bytes"""

    def __init__(
            self,
            reader: asyncio.StreamReader,
            writer: asyncio.StreamWriter,
            connector: 'BaseStreamConnector'
    ) -> None:
        self.reader, self.writer = reader, writer
        self.connector = connector
        self._send_queue = MessageQueue(DEFAULT_STREAM_HWM)

    def update(self, connection: Self):
        if not connection._send_queue.empty():
            raise ValueError('New connection\'s queue must be empty')
        self.reader, self.writer = connection.reader, connection.writer

    async def run_until_fail(
            self,
            request_received_cb: RequestReceivedCB,
    ):
        self._alive = True
        try:
            await self._run_connection(request_received_cb)
        except asyncio.IncompleteReadError:
            pass
        finally:
            self._alive = False
        return None

    async def send(self, metadata: Any, raw_data: bytes):
        # XXX: split big messages instead of error?
        if len(raw_data) > self._MAX_CHUNK_SIZE:
            raise ValueError('Data exceeds max chunk size (%d bytes)' %
                             self._MAX_CHUNK_SIZE)

        await self._send_queue.put(
            MessageFrame(json_dumpb(metadata), raw_data)
        )

    async def close(self):
        close_chunk = CloseNotifyFrame.to_bytes_with_header()
        with suppress(BrokenPipeError):
            self.writer.write(close_chunk)
        self.writer.close()
        with suppress(AssertionError):
            self.reader.feed_data(close_chunk)
        with suppress(BrokenPipeError):
            await self.writer.wait_closed()

    async def _run_connection(self, request_received_cb: RequestReceivedCB):
        is_closing = self.writer.is_closing
        readexactly = self.reader.readexactly
        header_size = Frame.length_size
        header_unpack = Frame.length_unpack_from

        write_task = asyncio.create_task(
            self._send_runner(),
            name=fmt_task_name('stream-connector-writer')
        )
        try:
            while not is_closing():
                header = await readexactly(header_size)
                chunk_len = header_unpack(header)[0]
                chunk = memoryview(await readexactly(chunk_len))

                frame = Frame.from_bytes(chunk)
                logger.debug('Received %r', frame)
                if isinstance(frame, MessageFrame):
                    request_received_cb(frame.metadata, frame.raw_data)
                elif isinstance(frame, CloseNotifyFrame):
                    self.writer.close()
                    return
                else:
                    logger.warning('wtf')
        except Exception as e:
            logger.info('Connection broken: %r', e)
            raise
        finally:
            write_task.cancel()
            await asyncio.wait([write_task])

    async def _send_runner(self):
        write = self.writer.write
        drain = self.writer.drain
        header_pack = Frame.length_pack
        send_queue_get = self._send_queue.get

        while True:
            frame = await send_queue_get()
            data = frame.to_bytes()

            write(
                header_pack(len(data)) +
                data
            )
            logger.debug('Sent %r', frame)

            # yield to loop, giving chance to pause protocol
            await asyncio.sleep(0)

            await drain()

    async def _do_handshake(
            self,
            handshaker: Handshaker,
    ):
        async def send_message(data):
            if NULL_CHAR in data or ETX_CHAR in data:
                raise ValueError('Handshake messages must NOT contain '
                                f'{NULL_CHAR!r} and {ETX_CHAR!r} characters')
            self.writer.write(data + NULL_CHAR)
            await self.writer.drain()

        async def recv_message() -> bytes:
            try:
                return (await self.reader.readuntil(NULL_CHAR))[:-1]
            except asyncio.IncompleteReadError as e:
                if (pos := e.partial.find(ETX_CHAR)) == -1:
                    raise
                raise HandshakeFail.loadb(e.partial[pos+1:])

        try:
            await self._run_handshaker(handshaker, send_message, recv_message)
            return True
        except HandshakeFail as fail:
            self.writer.write(ETX_CHAR + fail.dumpb())
            self.writer.close()
            await self.writer.wait_closed()
            return False


class BaseStreamConnector(BaseConnector):
    __slots__ = ()

    def _create_server_cb(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB
    ):
        async def callback(r: asyncio.StreamReader, w: asyncio.StreamWriter) -> None:
            connection = StreamConnection(r, w, self)
            if not await connection._do_handshake(handshaker):
                return
            client_connected_cb(connection)

        return callback

    @abstractmethod
    async def _open_connection(self) -> 'tuple[asyncio.StreamReader, asyncio.StreamWriter]':
        raise NotImplementedError

    async def client_connect(self, handshaker: Handshaker):
        r, w = await self._open_connection()
        connection = StreamConnection(r, w, self)
        try:
            await connection._do_handshake(handshaker)
        except HandshakeFail as fail:
            w.write(ETX_CHAR + fail.dumpb())
            w.close()
            await w.wait_closed()
            raise
        return connection


class TcpConnector(BaseStreamConnector):
    _TYPE = 'TcpConnector'

    __slots__ = ('host', 'port', 'ssl')

    def __init__(self, host: str, port: int, ssl: Optional[SSLContext] = None):
        self.host = host
        self.port = port
        self.ssl = ssl

    def repr_address(self) -> str:
        return f'{self.host}:{self.port}'

    async def server_start(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB,
    ):
        server = await asyncio.start_server(
            self._create_server_cb(handshaker, client_connected_cb),
            host=self.host,
            port=self.port,
            ssl=self.ssl
        )
        return BaseConnectorServer(server)

    async def _open_connection(self):
        return await asyncio.open_connection(
            host=self.host,
            port=self.port,
            ssl=self.ssl
        )

    def dump_state(self) -> str:
        return json.dumps({
            'type': self._TYPE,
            'host': self.host,
            'port': self.port,
            'ssl': bool(self.ssl)
        })

    @classmethod
    def from_state(cls, state: str, ssl: Optional[SSLContext] = None):
        data = json.loads(state)
        cls._check_dump(data)
        if data['ssl'] and not ssl:
            raise ValueError('SSL context was not provided, but connector state '
                             'demands it\'s presence')
        return cls(data['host'], data['port'], ssl)


class LocalConnector(BaseStreamConnector):
    """
    Connector for processes on same machine.

    Uses Named Pipes on Windows and Unix Domain Sockets on POSIX OS.
    """

    _TYPE = 'LocalConnector'

    __slots__ = ('_name',)

    @property
    def name(self):
        return self._name

    def __init__(self, name: str):
        """
        Args:
            name: Address of named pipe, or name of socket in tempdir.
              There is limit for name length, depending on running OS.
              Case insensitive (will be lowered).
        """
        if not local_connections.IS_AVAILABLE:
            raise FeatureNotAvailable(
                'Current platform does not provide named pipes or unix '
                'sockets. If there is some similar way of communication '
                'for local processes, please file an issue to communica\'s'
                'GitHub repository ("Bug Tracker" on PyPI page).'
            )
        self._name = name.lower()

    def repr_address(self) -> str:
        return local_connections.format_address(self._name)

    async def server_start(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB,
    ):
        return await local_connections.start_server(
            self._create_server_cb(handshaker, client_connected_cb),
            self._name
        )

    async def _open_connection(self):
        return await local_connections.open_connection(self._name)

    def dump_state(self) -> str:
        return json.dumps({'type': self._TYPE, 'name': self._name})

    @classmethod
    def from_state(cls, state: str):
        data = json.loads(state)
        cls._check_dump(data)
        return cls(data['name'])



from .connector import (
    DEFAULT_POOL,
    RmqServer,
    RmqConnector,
    RmqConnection,
    ClientCheckPolicy,
    ServerCheckPolicy,
    ConnectionCheckPolicy,
)


# XXX
__all__ = (
    'DEFAULT_POOL',
    'RmqServer',
    'RmqConnector',
    'RmqConnection',
    'ClientCheckPolicy',
    'ServerCheckPolicy',
    'ConnectionCheckPolicy',
)




import asyncio
import logging
import warnings
import threading
from uuid import uuid4
from typing import Any, Callable
from collections import defaultdict
from dataclasses import dataclass

from typing_extensions import Self

from communica.utils import read_accessor


try:
    import aiormq
    from yarl import URL
except ModuleNotFoundError:
    pass


logger = logging.getLogger('communica.connectors.rabbitmq')


class ChannelExpired(RuntimeError):
    pass


class ChannelMaxReached(Exception):
    pass


class PooledChannel:
    __slots__ = ('url', '_pool', '_chan', '_expired',
                 '_consumer_tag', '_release_callbacks')

    number: read_accessor[int] = read_accessor('_chan.number')
    explicitly_released: read_accessor[bool] = read_accessor('_expired')

    def __init__(
            self,
            url: 'URL',
            pool: 'ChannelPool',
            chan: 'aiormq.abc.AbstractChannel',
    ) -> None:
        self.url = url
        self._pool = pool
        self._chan = chan
        self._expired = False
        self._consumer_tag = None
        self._release_callbacks = []

    def __repr__(self) -> str:
        return (f'<{self.__class__.__name__} expired={self.expired} '
                f'ctag={self._consumer_tag} inner={self._chan!r}>')

    def __del__(self):
        if not self._expired:
            warnings.warn(f'Unreleased {self!r}', ResourceWarning, source=self)

    @property
    def expired(self) -> bool:
        return (self._expired or self._chan.is_closed)

    @property
    def inner(self) -> 'aiormq.abc.AbstractChannel':
        if self._expired:
            raise ChannelExpired(f'{self!r} is expired')
        if self._chan.is_closed:
            raise ChannelExpired(f'Underlying {self._chan!r} is closed')
        return self._chan

    def on_release(self, cb: Callable[[Self], None]):
        self._release_callbacks.append(cb)

    def publish(
            self,
            msg_type: str,
            body: bytes,
            exchange: str,
            routing_key: str,
            properties_kwargs: 'dict[str, Any]' = {}
    ):
        return self.inner.basic_publish(
            body,
            exchange=exchange,
            routing_key=routing_key,
            properties=aiormq.spec.Basic.Properties(  # pyright: ignore[reportUnboundVariable]
                message_type=msg_type,
                **properties_kwargs
            )
        )

    async def ack_message(self, message):
        await self._chan.basic_ack(message.delivery_tag)

    async def consume(self, queue: str, cb: 'aiormq.abc.ConsumerCallback', **kwargs):
        """
        Start queue consumer.
        One PooledChannel can have only one active consumer.
        """
        if self._consumer_tag:
            raise RuntimeError(f'{self!r} already consuming')
        self._consumer_tag = uuid4().hex
        return await self.inner.basic_consume(
            queue,
            cb,
            consumer_tag=self._consumer_tag,
            **kwargs
        )

    async def cancel_consume(self):
        if not self._consumer_tag:
            raise RuntimeError(f'{self!r} not consuming')
        if self._chan.is_closed:
            return
        coro = self._chan.basic_cancel(consumer_tag=self._consumer_tag)
        self._consumer_tag = None
        await coro

    async def queue_declare_and_bind(
            self,
            exchange: str,
            name: str = '',
            **declare_kwargs
    ):
        declare_ok = await self.inner.queue_declare(queue=name, **declare_kwargs)
        assert declare_ok.queue is not None, 'Incompatible AMQP server'
        queue_name = declare_ok.queue

        await self.inner.queue_bind(
            queue_name,
            exchange,
            routing_key=queue_name
        )
        return queue_name

    async def release(self):
        """
        Set .expired = true, restrict .inner access
        and return channel to pool. Can be called multiple times.
        """
        if self._expired:
            return
        self._expired = True

        for cb in self._release_callbacks:
            cb(self)

        if self._consumer_tag:
            try:
                await self.cancel_consume()
            except aiormq.exceptions.ChannelInvalidStateError:
                pass

        await self._pool._release_chan(self.url, self._chan)


class ChannelRoute:
    __slots__ = ('chan', '_exchange', '_routing_key')

    routing_key: read_accessor[str] = read_accessor('_routing_key')

    def __init__(
            self,
            chan: PooledChannel,
            exchange_name: str,
            routing_key: str
    ) -> None:
        self.chan = chan
        self._exchange = exchange_name
        self._routing_key = routing_key

    def publish(
            self,
            msg_type: str,
            body: bytes,
            properties_kwargs: 'dict[str, Any]' = {}
    ):
        return self.chan.publish(
            msg_type,
            body,
            exchange=self._exchange,
            routing_key=self._routing_key,
            properties_kwargs=properties_kwargs
        )


class ChannelPool:
    __slots__ = ('_loop_contexts', '_thread_lock')

    _thread_lock: threading.Lock
    _loop_contexts: 'dict[asyncio.AbstractEventLoop, _LoopContext]'

    @dataclass(frozen=True)
    class _LoopContext:
        locks: 'defaultdict[URL, asyncio.Lock]'
        connections: 'dict[URL, aiormq.abc.AbstractConnection]'
        free_channels: 'set[aiormq.abc.AbstractChannel]'
        release_waiters: 'set[asyncio.Future]'

    def __init__(self) -> None:
        self._thread_lock = threading.Lock()
        self._loop_contexts = {}

    def _check_loops(self):
        msg = '%r at 0x%02x is closed, but unclosed AMQP connection (%r) still '\
            'associated with %r. This may lead to weird bugs, as the broker '\
            'likely still considers that connection exists. Before closing '\
            'the asyncio loop you should call .close(url) method on the pool.\n%s'

        for loop, ctx in self._loop_contexts.items():
            if not loop.is_closed() or not ctx.connections:
                continue

            for conn in ctx.connections.values():
                if conn.is_closed:
                    continue

            from os import environ  # noqa
            from communica.connectors.rabbitmq import DEFAULT_POOL  # noqa

            note = ''
            if environ.get('PYTEST_CURRENT_TEST'):
                note += 'If you run this code within pytest, then either '\
                    'create new connector with new ChannelPool for each test, '\
                    'or call channel_pool.purge() after each test '\
                    '(e.g. in RmqConnector fixture finalization).\n'
            else:
                note += '(actually you should use only ONE event loop per ONE '\
                    'process, but whatever)\n'
            if self is DEFAULT_POOL:
                note += 'This is default pool. You can access it on '\
                    'communica.connectors.rabbitmq.DEFAULT_POOL. Also consider '\
                    'specifying your own pool in RmqConnector arguments.\n'

            logger.warning(msg, loop, id(loop), conn, self, note.strip())

    def _get_current_context(self) -> _LoopContext:
        loop = asyncio.get_running_loop()

        context = self._loop_contexts.get(loop)
        if context is not None:
            return context

        with self._thread_lock:
            self._check_loops()

            if loop not in self._loop_contexts:
                self._loop_contexts[loop] = self._LoopContext(
                    locks=defaultdict(lambda: asyncio.Lock()),
                    connections=dict(),
                    free_channels=set(),
                    release_waiters=set()
                )
        return self._loop_contexts[loop]

    async def acquire_chan(self, url: 'URL') -> PooledChannel:
        context = self._get_current_context()

        async with context.locks[url]:
            conn = context.connections.get(url)
            if conn is None or conn.is_closed:
                conn = await aiormq.connect(url)
                context.connections[url] = conn

            if len(conn.channels) >= conn.connection_tune.channel_max:
                logger.warning(
                    'Negotiated channel limit for %r reached. Channel '
                    'will not be acquired until other channel released.',
                    url.with_password(None)
                )
                fut = asyncio.get_running_loop().create_future()
                context.release_waiters.add(fut)
                await fut

            if context.free_channels:
                rmq_chan = context.free_channels.pop()
                assert not rmq_chan.is_closed, 'Channel in pool was unexpectedly closed.'
            else:
                rmq_chan = await conn.channel()

            logger.debug('%r acquired from %r (%r)',
                         rmq_chan, self, url.with_password(None))
            return PooledChannel(url, self, rmq_chan)

    async def _release_chan(self, url: 'URL', rmq_chan: 'aiormq.abc.AbstractChannel'):
        logger.debug('%r released from %r (%r)',
                     rmq_chan, self, url.with_password(None))
        context = self._get_current_context()

        async with context.locks[url]:
            if not rmq_chan.is_closed:
                context.free_channels.add(rmq_chan)

            while context.release_waiters:
                waiter = context.release_waiters.pop()
                if not waiter.done():
                    waiter.set_result(None)
                    break

    async def close(self, url: 'URL') -> bool:
        """
        Close connection for given url.
        If there is no connection in pool, returns False.

        This does not prevent opening of the new connections.
        """
        context = self._get_current_context()
        async with context.locks[url]:
            while context.release_waiters:
                waiter = context.release_waiters.pop()
                if not waiter.done():
                    waiter.set_exception(ChannelMaxReached())

            conn = context.connections.pop(url, None)
            if conn is None:
                return False
            if conn.is_opened:
                await conn.close()
            return True

    async def purge(self):
        """
        Close all connections associated with current event loop.
        """
        context = self._get_current_context()
        connections = list(context.connections.values())

        for conn in connections:
            await self.close(conn.url)



try:
    import adaptix

    _HAVE_ADAPTIX = True
except ModuleNotFoundError:
    _HAVE_ADAPTIX = False

from typing import Any, Type, Generic, TypeVar

from communica.utils import ByteSeq
from communica.exceptions import FeatureNotAvailable
from communica.serializers import BaseSerializer
from communica.serializers.json import JsonSerializer


TReq = TypeVar('TReq')
TResp = TypeVar('TResp')

DEFAULT_JSON_SERIALIZER = JsonSerializer()


class AdaptixSerializer(BaseSerializer, Generic[TReq, TResp]):
    """
    This serializer passes data to `adaptix.Retort`

    If you specify `request_model` as A and `response_model` as B:
    - On request side, it dumps A (out) and loads B (in)
    - On response side, it loads A (in) and dumps B (out)

    So, validation ensured on both sides.
    If the same serializer defined on both sides, of course.
    """

    __slots__ = ('_retort', '_req_model', '_resp_model', '_json')

    # XXX: один из вариантов
    # FROM_HANDLER = специальное значение
    # если регистратор ручки видит это значение вместо сериалайзера,
    # он берет ручку и через какой-нибудь метод собирает сериалайзер
    # по сигнатуре ручки

    def __init__(
            self,
            request_model: 'Type[TReq] | None',
            response_model: 'Type[TResp] | None' = None,
            retort: 'adaptix.Retort | None' = None,
            json_serializer: 'JsonSerializer | None' = None,
    ) -> None:
        """
            Args:
                request_model: Type of object, which will be sent
                    from requester to responder.
                    None means no validation/conversion performed.
                response_model: Type of object, which will be sent
                    from responder to requester.
                    None means no validation/conversion performed.
        """

        if not _HAVE_ADAPTIX:
            raise FeatureNotAvailable(
                'AdaptixSerializer requires adaptix library. '
                'Install communica with [adaptix] extra.'
            )

        self._json = json_serializer or DEFAULT_JSON_SERIALIZER
        self._retort = retort or adaptix.Retort()  # pyright: ignore[reportPossiblyUnboundVariable]
        self._req_model = request_model or Any
        self._resp_model = response_model or Any

    def load(self, raw_data: ByteSeq) -> TReq:
        'Load data as request_model'
        return self._retort.load(
            self._json.load(raw_data),
            self._req_model
        )  # pyright: ignore[reportReturnType]  # until typevar defaults available

    def dump(self, data: TResp) -> bytes:
        'Dump data as response_model'
        return self._json.dump(
            self._retort.dump(data, self._resp_model)
        )

    def client_load(self, raw_data: ByteSeq) -> TResp:
        'Load data as response_model'
        return self._retort.load(
            self._json.load(raw_data),
            self._resp_model
        )  # pyright: ignore[reportReturnType]

    def client_dump(self, data: TReq) -> bytes:
        'Dump data as request_model'
        return self._json.dump(
            self._retort.dump(data, self._req_model)
        )



from communica.entities.route import RouteTable, RouteClient, RouteServer
from communica.entities.simple import SimpleClient, SimpleServer


__all__ = (
    'SimpleClient',
    'SimpleServer',
    'RouteClient',
    'RouteServer',
    'RouteTable',
)



from .base import BaseConnector
from .stream import TcpConnector, LocalConnector
from .rabbitmq import RmqConnector


__all__ = (
    'TcpConnector',
    'RmqConnector',
    'LocalConnector',
    'BaseConnector'
)



import asyncio
from uuid import uuid4
from typing import Any, TypedDict

from communica.utils import TaskSet, logger, fmt_task_name
from communica.exceptions import ReqError
from communica.serializers import BaseSerializer, default_serializer
from communica.entities.base import BaseClient, SyncHandlerType, AsyncHandlerType
from communica.connectors.base import (
    BaseConnector,
)
from communica.entities.simple import (
    RequestType,
    ReqRepClient,
    ReqRepServer,
    RequestHandler,
    ReqRepMessageFlow,
)


class PubSubClient(BaseClient): ...



import logging
import itertools
import threading
from random import gauss
from typing import Generic, TypeVar
from asyncio import Task, sleep, current_task, get_running_loop
from inspect import isclass, ismethod, isfunction
from operator import attrgetter
from traceback import format_exception
from collections import deque, defaultdict

from typing_extensions import TypeAlias


_TV = TypeVar('_TV')
ByteSeq: TypeAlias = 'bytes | memoryview | bytearray'

logger = logging.getLogger('communica')
_task_counters = defaultdict(lambda: itertools.count())

ETX_CHAR = b'\x04'
NULL_CHAR = b'\x00'

UINT32MAX = 2**32 - 1
UINT16MAX = 2**16 - 1
INT32RANGE = (-2**31, 2**31 - 1)


_HasLoopMixin_lock = threading.Lock()


class read_accessor(Generic[_TV]):
    """
    ```
    class Foo:
        url: read_accessor[str] = read_accessor('_url')
    ```
    is equivalent for
    ```
    class Foo:
        @property
        def url(self) -> str:
            return self._url
    ```
    """
    __slots__ = ('_getter',)

    def __init__(self, path: str) -> None:
        self._getter = attrgetter(path)

    def __get__(self, instance, owner=None) -> _TV:
        return self._getter(instance)


# can be replaced with asyncio.mixins._LoopBoundMixin,
# but it's not in public API
class HasLoopMixin:
    _bound_loop = None

    def _get_loop(self):
        loop = get_running_loop()

        if self._bound_loop is None:
            with _HasLoopMixin_lock:
                if self._bound_loop is None:
                    self._bound_loop = loop
        if loop is not self._bound_loop:
            raise RuntimeError(
                f'{self!r} is bound to a different event loop. '
                'Consider using one event loop per process, '
                'or at least do not use one asyncio object '
                'in different loops and/or threads.'
            )
        return loop


class TaskSet(HasLoopMixin):
    tasks: 'set[Task]'

    def __init__(self) -> None:
        self.tasks = set()

    def create_task(self, coro, *, name=None) -> Task:
        task = self._get_loop().create_task(coro, name=name)
        task.add_done_callback(self.tasks.discard)
        self.tasks.add(task)
        return task

    def create_task_with_exc_log(self, coro, *, name=None):
        task = self.create_task(coro, name=name)
        task.add_done_callback(exc_log_callback)
        return task

    def cancel(self):
        "Cancel all tasks except current"
        cur_task = current_task()
        for task in self.tasks:
            if task is not cur_task:
                task.cancel()


class MessageQueue(HasLoopMixin, Generic[_TV]):
    __slots__ = ('_queue', '_max_items', '_get_waiter', '_put_waiters')

    def __init__(self, max_items: int) -> None:
        self._queue = deque()
        self._max_items = max_items
        self._get_waiter = None
        self._put_waiters = deque()

    def empty(self):
        return (len(self._queue) == 0)

    def full(self):
        return (len(self._queue) >= self._max_items)

    def set_max_items(self, max_items: int):
        self._max_items = max_items

    def put_back(self, item: _TV):
        self._queue.appendleft(item)

    async def put(self, item: _TV):
        if len(self._queue) >= self._max_items:
            waiter = self._get_loop().create_future()
            self._put_waiters.append(waiter)
            await waiter

        if self._get_waiter:
            if not self._get_waiter.done():
                self._get_waiter.set_result(True)
            self._get_waiter = None

        self._queue.append(item)

    async def get(self) -> _TV:
        if not self._queue:
            if self._get_waiter and not self._get_waiter.done():
                raise RuntimeError('Duplicate get from queue')

            waiter = self._get_loop().create_future()
            self._get_waiter = waiter
            await self._get_waiter

        while self._put_waiters:
            if not (waiter := self._put_waiters.popleft()).done():
                waiter.set_result(True)
                break

        return self._queue.popleft()


class BackoffDelayer:
    __slots__ = ('start_delay', 'max_delay', 'factor', 'jitter', '_delay')

    def __init__(
            self,
            start_delay: float,
            max_delay: float,
            factor: float,
            jitter: float
    ) -> None:
        assert factor > 1, 'factor less or equal 1'
        assert max_delay > start_delay, 'max delay less or equal to start delay'
        self.factor = factor
        self.jitter = jitter

        self._delay = start_delay
        self.max_delay = max_delay
        self.start_delay = start_delay

    def reset(self):
        self._delay = self.start_delay

    async def wait(self):
        await sleep(self._delay)
        new_delay = gauss(
            min(self.max_delay, (self._delay * self.factor)),
            self.jitter
        )
        if new_delay > self._delay:
            self._delay = new_delay


def exc_log_callback(task: Task):
    if not task.cancelled() and (exc := task.exception()):
        tb = format_exception(type(exc), exc, exc.__traceback__)
        logger.warning('Uncaught exception in %r:\n%s', task, '\n'.join(tb))


def iscallable(obj, _prev=None):
    if isfunction(obj):
        return True
    if isclass(obj):
        return False
    if ismethod(obj):
        return True
    if obj is _prev:
        return False
    return iscallable(getattr(obj, '__call__', None), obj)


def cycle_range(start: int, stop: int):
    if stop <= start:
        ValueError('stop should be greater than start')

    i = start
    while True:
        yield i
        if i >= stop:
            i = 0
        else:
            i += 1


def fmt_task_name(name: str) -> str:
    return f'communica-{name}-{next(_task_counters[name])}'



import asyncio
from uuid import uuid4
from typing import Any, Iterable, TypedDict

from typing_extensions import Self

from communica.utils import ByteSeq, TaskSet, logger, fmt_task_name
from communica.exceptions import ReqError, RouteOverrideError
from communica.serializers import BaseSerializer, default_serializer
from communica.entities.base import HandlerType
from communica.connectors.base import (
    BaseConnector,
)
from communica.entities.simple import (
    RequestType,
    ReqRepClient,
    ReqRepServer,
    RequestHandler,
    ReqRepMessageFlow,
)


__all__ = (
    'RouteClient',
    'RouteServer'
)


class Metadata(TypedDict):
    id: str
    type: RequestType
    route: str


class RouteTable:
    """
    This object allows you to register routes in global variable
    and then use them when creating RouteServer or RouteClient.
    """

    __slots__ = ('route_map', 'name')

    route_map: 'dict[str, tuple[RequestHandler, BaseSerializer]]'

    def __init__(self, name: 'str | None' = None) -> None:
        self.name = name or 'unnamed'
        self.route_map = {}

    def __repr__(self):
        clsname = self.__class__.__name__
        return f'<{clsname} name={self.name}>'

    def add_route(
            self,
            route: str,
            handler: HandlerType,
            serializer: 'BaseSerializer | None'
    ):
        if route in self.route_map:
            raise RouteOverrideError(f'Route {route!r} already defined')

        if serializer is None:
            serializer = default_serializer
        if not isinstance(handler, RequestHandler):
            handler = RequestHandler(handler)
        self.route_map[route] = (handler, serializer)

    def handler(
            self,
            route: str,
            serializer: 'BaseSerializer | None' = None
    ):
        """
        Register handler for route.

        Args:
            route: Handler identifier.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        def decorator(handler: HandlerType):
            self.add_route(route, handler, serializer)
            if isinstance(handler, RequestHandler):
                return handler.endpoint
            return handler

        return decorator

    def __getitem__(self, __key: str) -> 'tuple[RequestHandler, BaseSerializer]':
        return self.route_map[__key]

    @classmethod
    def _join_route_tables(
            cls,
            anothers: 'Iterable[Self | None] | Self | None'
    ) -> Self:
        """
        Create new table with routes from anothers.
        """
        result_table = cls()

        def add_table(table: 'RouteTable | None'):
            if table is None:
                return
            for route, value in table.route_map.items():
                if route in result_table.route_map:
                    break
                result_table.route_map[route] = value
            else:
                return

            for antable in anothers:  # pyright: ignore
                if route not in antable.route_map:
                    continue
                raise RouteOverrideError(
                    f'Can\'t join {antable} with {table}: '
                    f'route {route!r} already defined.'
                )

        if anothers is None or isinstance(anothers, RouteTable):
            add_table(anothers)
            if anothers is not None:
                result_table.name = anothers.name
        else:
            for table in anothers:
                add_table(table)
            result_table.name = 'joined'
        return result_table


class RouteMessageFlow(ReqRepMessageFlow):
    __slots__ = ('routes', 'fallback_task_set', '_response_serializers',)

    routes: RouteTable
    fallback_task_set: TaskSet

    _response_serializers: 'dict[str, BaseSerializer]'

    def __init__(self, route_table: RouteTable):
        super().__init__()
        self.routes = route_table
        self.fallback_task_set = TaskSet()
        self._response_serializers = {}

    def dispatch(self, metadata: Metadata, raw_data: ByteSeq):
        if metadata['type'] < RequestType.RESP_OK:
            try:
                route_handle = self.routes[metadata['route']]
            except KeyError:
                route_handle = None

            self.task_set.create_task_with_exc_log(
                self.handle_request(route_handle, metadata, raw_data),
                name=fmt_task_name('route-request-handler')
            )
        else:
            try:
                serializer = self._response_serializers[metadata['route']]
            except KeyError:  # pragma: no cover
                # this is possible after program restart.
                # just log and skip, cause we don't have a routine,
                # waiting for this response
                # XXX: should it be logged?
                logger.warning('Got response with not requested '
                              f'route ({metadata["route"]})')
                return
            self._handle_response(serializer, metadata, raw_data)

    async def handle_request(
            self,
            route_handle: 'tuple[RequestHandler, BaseSerializer] | None',
            req_meta: Metadata,
            raw_data: ByteSeq
        ):
        if route_handle is not None:
            handler, serializer = route_handle
            resp_meta, resp_data = await self._handle_request(
                handler, serializer, req_meta, raw_data
            )
        else:
            serializer = default_serializer
            resp_meta = req_meta.copy()
            resp_meta['type'] = RequestType.RESP_ERR_REQUESTER
            resp_data = \
                ReqError(f'Route "{req_meta["route"]}" is not defined').to_dict()

        await self._send_response(req_meta, resp_meta, resp_data, serializer)

    async def request(
            self,
            route: str,
            serializer: 'BaseSerializer | None',
            data: Any
    ) -> bytes:
        # TODO: timeout
        request_id, fut = self._create_response_waiter()

        serializer = serializer or default_serializer
        if (saved_serializer := self._response_serializers.get(route)) is None:
            self._response_serializers[route] = serializer
        elif saved_serializer is not serializer:
            raise TypeError('You should always use the same serializer for '
                           f'the same route ({saved_serializer!r} '
                            'has already been used earlier)')

        await self._connection.send(
            Metadata(id=request_id, type=RequestType.REQ_REP, route=route),
            serializer.client_dump(data)
        )

        return await fut

    async def throw(
            self,
            route: str,
            serializer: 'BaseSerializer | None',
            data: Any
    ) -> None:
        await self._connection.send(
            Metadata(id='', type=RequestType.REQ_THROW, route=route),
            (serializer or default_serializer).client_dump(data)
        )


class RouteClient(ReqRepClient[RouteMessageFlow]):
    """
    Pair for RouteServer.

    Handler searched by exact match of route string.
    """

    __slots__ = ()

    @property
    def routes(self):
        return self._flow.routes

    def __init__(
            self,
            connector: BaseConnector,
            client_id: 'str | None' = None,
            route_table: 'Iterable[RouteTable] | RouteTable | None' = None
    ) -> None:
        super().__init__(connector)

        self._flow = RouteMessageFlow(
            RouteTable._join_route_tables(route_table)
        )
        self._run_task = None
        self._client_id = client_id or uuid4().hex

    def route_handler(self, *args, **kwargs):
        "Deprecated"
        return self.routes.handler(*args, **kwargs)

    async def request(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None
    ) -> Any:
        """
        Send request, wait response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        if not self._connected_event.is_set():
            await self._connected_event.wait()
        return await self._flow.request(route, serializer, data)

    async def throw(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None
    ) -> None:
        """
        Send request without waiting response.
        However, this method will block until data is sent.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer.
        """
        if not self._connected_event.is_set():
            await self._connected_event.wait()
        return await self._flow.throw(route, serializer, data)


class RouteServer(ReqRepServer[RouteMessageFlow]):
    """
    Pair for RouteClient.

    Handler searched by exact match of route string.
    """

    __slots__ = ('_route_table',)

    _route_table: RouteTable
    _known_clients: 'dict[str, RouteMessageFlow | asyncio.Future[RouteMessageFlow]]'

    @property
    def routes(self):
        return self._route_table

    def __init__(
            self,
            connector: BaseConnector,
            route_table: 'RouteTable | None' = None
    ) -> None:
        self.connector = connector
        self._known_clients = {}
        self._client_connected = asyncio.Event()
        self._client_conn_runners = {}

        self._route_table = RouteTable._join_route_tables(route_table)

    def _create_new_flow(self) -> RouteMessageFlow:
        return RouteMessageFlow(self._route_table)

    def route_handler(self, *args, **kwargs):
        "Deprecated"
        return self.routes.handler(*args, **kwargs)

    async def request(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None,
            client_id: 'str | None' = None
    ) -> Any:
        """
        Send request, wait response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random connected client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer. Once serializer was used for
                request, no other serializer can be used for same route.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.request(route, serializer, data)

    async def throw(
            self,
            route: str,
            data: Any,
            serializer: 'BaseSerializer | None' = None,
            client_id: 'str | None' = None
    ) -> None:
        """
        Send request without waiting response.

        Args:
            route: Handler identifier.
            client_id: If omitted or None, random connected client will be chosen.
            serializer: Instance of :obj:`communica.serializers.BaseSeralizer`.
                Defaults to JsonSerializer. Once serializer was used for
                request, no other serializer can be used for same route.
        """
        flow = await self._get_client_flow(client_id)
        return await flow.throw(route, serializer, data)



from .entities import (
    RouteTable,
    RouteClient,
    RouteServer,
    SimpleClient,
    SimpleServer,
)
from .connectors import (
    RmqConnector,
    TcpConnector,
    LocalConnector,
)
from .serializers import (
    JsonSerializer,
    AdaptixSerializer,
)


__all__ = (
    'RouteTable',
    'RouteClient',
    'RouteServer',
    'SimpleClient',
    'SimpleServer',
    'RmqConnector',
    'TcpConnector',
    'LocalConnector',
    'JsonSerializer',
    'AdaptixSerializer',
)



import sys
import asyncio
from abc import ABC, abstractmethod
from typing import Any, Callable, Protocol, Awaitable, AsyncGenerator

from communica.utils import logger
from communica.serializers.json import json_dumpb, json_loadb


HandshakeGen = AsyncGenerator['dict | HandshakeOk', dict]


class Handshaker(Protocol):
    def __call__(self, connection: 'BaseConnection') -> HandshakeGen: ...


class ClientConnectedCB(Protocol):
    def __call__(self, connection: 'BaseConnection'): ...


class RequestReceivedCB(Protocol):
    def __call__(self, metadata: Any, raw_data: 'bytes | memoryview'): ...


class HandshakeOk: ...


class HandshakeFail(Exception):
    def __init__(self, reason: str) -> None:
        self.reason = reason

    def dumpb(self):
        obj = {'reason': self.reason}
        return json_dumpb(obj)

    @classmethod
    def loadb(cls, data: bytes):
        obj = json_loadb(data)
        return cls(obj['reason'])


class BaseConnection(ABC):
    __slots__ = ('_alive', '_handshake_result')

    @property
    def is_alive(self) -> bool:
        try:
            return self._alive
        except AttributeError:
            return False

    def get_handshake_result(self) -> HandshakeOk:
        """
        Get result of recent handshake.

        Handshake should be performed
        by connection only once right after connect.

        Raises:
            RuntimeError: if called more than once
        """
        try:
            result = self._handshake_result
        except AttributeError:
            raise RuntimeError('Handshake not completed yet') from None
        if result is None:
            raise RuntimeError('Handshake result already has been received')
        return result

    async def _run_handshaker(
            self,
            handshaker: Handshaker,
            send_message: Callable[[bytes], Awaitable],
            recv_message: Callable[[], Awaitable[bytes]]
    ) -> 'HandshakeOk':
        handshake_gen = handshaker(self)

        async def send_to_gen(data):
            if data is not None:
                data = json_loadb(data)
            gen_return = await handshake_gen.asend(data)
            if isinstance(gen_return, dict):
                return json_dumpb(gen_return)
            elif isinstance(gen_return, HandshakeOk):
                if hasattr(self, '_handshake_result'):
                    raise RuntimeError('Handshake repeated on same connection')
                self._handshake_result = gen_return
                return gen_return
            else:
                raise StopAsyncIteration

        try:
            # Первый asend должен быть с None
            data_or_result = await send_to_gen(None)
            if isinstance(data_or_result, HandshakeOk):
                return data_or_result

            while True:
                await send_message(data_or_result)

                received_data = await recv_message()

                data_or_result = await send_to_gen(received_data)
                if isinstance(data_or_result, HandshakeOk):
                    return data_or_result
        except HandshakeFail:
            raise
        except StopAsyncIteration:
            raise RuntimeError('Handshake generator must return bytes, '
                               'HandshakeOk or HandshakeFail') from None
        except Exception as error:
            logger.exception('Handshake failed: %r raised %r', handshaker, error)
            raise
        finally:
            await handshake_gen.aclose()

    @abstractmethod
    def update(self, connection) -> None:
        """
        Update information, required for connection use.
        This method invoked after reconnect on the old
        connection object with new connection as a parameter and
        then old connection might be used for communication.

        Must be called before run_until_fail.
        """
        raise NotImplementedError

    @abstractmethod
    async def send(self, metadata: Any, raw_data: bytes):
        raise NotImplementedError

    @abstractmethod
    async def run_until_fail(
            self,
            request_received_cb: RequestReceivedCB
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError


class BaseConnectorServer:
    def __init__(self, server: asyncio.AbstractServer) -> None:
        self.server = server  # pyright: ignore[reportGeneralTypeIssues]

    def close(self) -> None:
        self.server.close()

    def is_serving(self) -> bool:
        return self.server.is_serving()

    if sys.version_info >= (3, 13):
        async def wait_closed(self):
            self.server.close_clients()
            await self.server.wait_closed()
    else:
        async def wait_closed(self):
            return


class BaseConnector(ABC):
    """Making connections"""
    _TYPE: str

    __slots__ = ()

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.repr_address()})'

    @abstractmethod
    def repr_address(self) -> str:
        raise NotImplementedError

    @classmethod
    def _check_dump(cls, dump: 'dict[str, Any]'):
        if cls._TYPE != dump['type']:
            raise ValueError('Wrong connector type: passed string was '
                            f'constructed with "{dump["type"]}", '
                            f'current is "{cls._TYPE}"')

    @abstractmethod
    async def server_start(
            self,
            handshaker: Handshaker,
            client_connected_cb: ClientConnectedCB,
    ) -> BaseConnectorServer:
        raise NotImplementedError

    @abstractmethod
    async def client_connect(self, handshaker: Handshaker) -> BaseConnection:
        raise NotImplementedError



# pyright: reportAttributeAccessIssue=none

import sys
import asyncio

from communica.connectors.base import BaseConnectorServer


IS_AVAILABLE = True


if sys.platform == 'win32':
    class PipeConnectorServer(BaseConnectorServer):
        def is_serving(self):
            return not self.server.closed()

        async def wait_closed(self):
            return


    async def open_connection(
            address: str
    ) -> 'tuple[asyncio.StreamReader, asyncio.StreamWriter]':
        loop = asyncio.get_event_loop()
        if not hasattr(loop, 'create_pipe_connection'):
            raise TypeError('Named pipes are not supported on current platform')

        reader = asyncio.StreamReader(loop=loop)
        protocol = asyncio.StreamReaderProtocol(reader, loop=loop)
        transport, _ = await loop.create_pipe_connection(
            lambda: protocol, format_address(address))
        writer = asyncio.StreamWriter(transport, protocol, reader, loop)
        return reader, writer

    async def start_server(
            client_connected_cb, address: str
    ) -> BaseConnectorServer:
        loop = asyncio.get_event_loop()
        if not hasattr(loop, 'start_serving_pipe'):
            raise TypeError('Named pipes are not supported on current platform')

        def factory():
            reader = asyncio.StreamReader(loop=loop)
            protocol = asyncio.StreamReaderProtocol(
                reader, client_connected_cb, loop=loop)
            return protocol

        servers = await loop.start_serving_pipe(factory, format_address(address))
        return PipeConnectorServer(servers[0])

    def format_address(address: str) -> str:
        return '\\\\.\\pipe\\' + 'communica.' + address

elif hasattr(asyncio, 'open_unix_connection'):
    import os
    import os.path
    import tempfile

    SOCK_DIR = os.path.join(tempfile.gettempdir(), 'communica')
    if not os.path.exists(SOCK_DIR):
        os.makedirs(SOCK_DIR)

    async def open_connection(
            address: str
    ) -> 'tuple[asyncio.StreamReader, asyncio.StreamWriter]':
        sock_path = format_address(address)
        return await asyncio.open_unix_connection(sock_path)

    async def start_server(
            client_connected_cb, address: str
    ) -> BaseConnectorServer:
        sock_path = format_address(address)
        server = await asyncio.start_unix_server(client_connected_cb, sock_path)
        return BaseConnectorServer(server)

    def format_address(address: str) -> str:
        return os.path.join(SOCK_DIR, address + '.sock')

else:
    IS_AVAILABLE = False
